using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using Pk2 = PICkit2V2.PICkitFunctions;
using KONST = PICkit2V2.Constants;
using UTIL = PICkit2V2.Utilities;
using ImportExportHex = PICkit2V2.ImportExportHex;
using Constants = PICkit2V2.Constants;
using Pk2BootLoader = PICkit2V2.Pk2BootLoader;
using DeviceData = PICkit2V2.DeviceData;

namespace QwikProgram2v2
{
    // State enumeration
    enum QP2State
    {
        ready, busy
    }

    // Task enumeration
    enum QP2Task
    {
        detect, detectAutoLoad, downloadHex, downloadFW
    }

    public partial class QwikProgram2Form : Form
    {
        // Background worker
        private BackgroundWorker mBW;

        // State
        private QP2State state;
        private string hexFilename, contentionMsg;
        private bool picDetected, pickitDetected, needFirmwareUpdate;
        private uint debugVector;
        private bool debugFW = false;

        // Status label
        private ToolStripLabel labelStatus;

        // Form initialization
        public QwikProgram2Form()
        {
            InitializeComponent();

            // Initialize the background worker
            InitBW();

            // Initialize GUI stuff
            this.labelStatus = new ToolStripLabel();
            statusStrip.Items.Add(labelStatus);
            progressBar.Value = 0;
            progressBar.Maximum = 100;

            // Initialize the state variables
            state = QP2State.ready;
            picDetected = pickitDetected = needFirmwareUpdate = false;
            hexFilename = contentionMsg = "";
            debugVector = 0x0000;

            // Initialize programming voltage.
            Pk2.SetVDDVoltage(5.0F);

            // Load the device file
            if (!loadDeviceFile())
            {
                MessageBox.Show("Invalid device file.\nPlease reinstall QwikProgram 2.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            // Detect everything
            state = QP2State.busy;
            updateGUI();
            mBW.RunWorkerAsync(QP2Task.detectAutoLoad);

            // Start the regular "poll" timer
            timer.Start();
        }

        // Initialize the background worker
        private void InitBW()
        {
            mBW = new BackgroundWorker();
            mBW.RunWorkerCompleted +=
                new RunWorkerCompletedEventHandler(mBW_RunWorkerCompleted);
            mBW.ProgressChanged
                += new ProgressChangedEventHandler(mBW_ProgressChanged);
            mBW.DoWork += new DoWorkEventHandler(mBW_DoWork);
            mBW.WorkerReportsProgress = true;
            mBW.WorkerSupportsCancellation = true;
        }

        // Completion event
        private void mBW_RunWorkerCompleted(object sender,
                                    RunWorkerCompletedEventArgs e)
        {
            state = QP2State.ready;

            updateGUI();
        }

        // Progress event
        private void mBW_ProgressChanged(object sender,
                                        ProgressChangedEventArgs e)
        {
            progressBar.Value = e.ProgressPercentage;
            labelStatus.Text = (string)e.UserState;

            updateGUI();
        }

        // On worker thread! Only method with real logic
        private void mBW_DoWork(object sender, DoWorkEventArgs e)
        {
            int digits = (int)e.Argument;
            BackgroundWorker bw = (BackgroundWorker)sender;
            QP2Task task = (QP2Task)e.Argument;

            switch (task)
            {
                case QP2Task.detectAutoLoad:
                    detectEverything(bw, true);
                    break;

                case QP2Task.detect:
                    detectEverything(bw, false);
                    break;

                case QP2Task.downloadHex:
                    deviceWrite(bw);
                    break;

                case QP2Task.downloadFW:
                    downloadNewFirmware(bw);
                    break;
            }
        }

        // Open hex file menu item
        private void loadHexFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (state == QP2State.ready && picDetected && pickitDetected)
                openHexFileDialog.ShowDialog();

            updateGUI();
        }

        // Exit menu item
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        // Usage guide menu item
        private void usageGuideToolStripMenuItem_Click(object sender, EventArgs e)
        {
            System.Diagnostics.Process.Start("help.html");
        }

        // Write button
        private void buttonWrite_Click(object sender, EventArgs e)
        {
            if (state != QP2State.ready)
                return;

            state = QP2State.busy;
            updateGUI();

            // Download QwikBug
            if (pickitDetected && picDetected &&
                hexFilename.Length > 0 && contentionMsg.Length == 0)
                mBW.RunWorkerAsync(QP2Task.downloadHex);

           // Download firmware
            else if (needFirmwareUpdate)
                mBW.RunWorkerAsync(QP2Task.downloadFW);
        }


        // Selected a file
        private void openHexFileDialog_FileOk(object sender, CancelEventArgs e)
        {
            if (state == QP2State.ready && picDetected && pickitDetected)
                importHexAndVectorFiles(openHexFileDialog.FileName);

            updateGUI();
        }

        // Update state
        private void detectEverything(BackgroundWorker bw, bool autoLoad)
        {
            bw.ReportProgress(0, "Searching for PICkit 2...");
            detectPICkit2(bw);

            if (pickitDetected)
            {
                bw.ReportProgress(50, "Searching for PIC...");
                checkForPowerErrors();
                detectPIC();

                if (picDetected)
                {
                    Pk2.ResetBuffers();
                    if (autoLoad)
                        importHexAndVectorFiles("qb" + Pk2.DevFile.PartsList[Pk2.ActivePart].PartName + ".hex");
                }
                else
                {
                    hexFilename = "";
                    Pk2.ResetBuffers();
                }
            }

            // Send progress update
            if (contentionMsg.Length > 0)
                bw.ReportProgress(100, contentionMsg);
            else if (needFirmwareUpdate)
                bw.ReportProgress(100, "Firmware update required.  Click Update.");
            else
            {
                string msg;
                if (pickitDetected)
                {
                    msg = "PICkit 2 v" + Pk2.FirmwareVersion;
                    if (picDetected)
                        msg += ", " + Pk2.DevFile.PartsList[Pk2.ActivePart].PartName
                            + " rev. " + Pk2.ActiveRev;
                }
                else
                    msg = "PICkit 2 not";
                msg += " detected.";

                bw.ReportProgress(100, msg);
            }
        }

        // Check for power supply conflicts
        private void checkForPowerErrors()
        {
            KONST.PICkit2PWR result = Pk2.PowerStatus();
            if (result == KONST.PICkit2PWR.vdderror)
                contentionMsg = "Warning: Vdd contention detected.";
            else if (result == KONST.PICkit2PWR.vpperror)
                contentionMsg = "Warning: Vss contention detected.";
            else if (result == KONST.PICkit2PWR.vddvpperrors)
                contentionMsg = "Warning: Vdd and Vss contention detected. ";
            else
                contentionMsg = "";
        }

        // Update the user interface
        private void updateGUI()
        {
            // Write button, import hex menu
            if (state != QP2State.ready)
                buttonWrite.Enabled = loadHexFileToolStripMenuItem.Enabled =
                    eraseToolStripMenuItem.Enabled = detectToolStripMenuItem.Enabled = false;
            else
            {
                if (needFirmwareUpdate)
                {
                    buttonWrite.Enabled = true;
                    loadHexFileToolStripMenuItem.Enabled = false;
                    detectToolStripMenuItem.Enabled = false;
                    eraseToolStripMenuItem.Enabled = false;
                    buttonWrite.Text = "Update";
                }
                else
                {
                    buttonWrite.Text = "Write";
                    detectToolStripMenuItem.Enabled = true;
                    if (picDetected && pickitDetected)
                    {
                        loadHexFileToolStripMenuItem.Enabled = true;
                        buttonWrite.Enabled = hexFilename.Length > 0;
                        eraseToolStripMenuItem.Enabled = true;
                    }
                    else
                    {
                        buttonWrite.Enabled = loadHexFileToolStripMenuItem.Enabled = false;
                        eraseToolStripMenuItem.Enabled = false;
                    }
                }
            }

            // Hex filename label
            if (hexFilename.Length > 0)
                labelHexFilename.Text = shortenHex(hexFilename);
            else
                labelHexFilename.Text = "(no hex file loaded)";

            // Update window and leave
            Update();
        }

        // Load the device file
        private bool loadDeviceFile()
        {
            if (Pk2.ReadDeviceFile())
            {
                if (Pk2.DevFile.Info.Compatibility < KONST.DevFileCompatLevel)
                    return false;
                if (Pk2.DevFile.Info.Compatibility > KONST.DevFileCompatLevel)
                    return false;

                return true;
            }
            return false;
        }

        // Import a hex file (and debug vector file, too)
        private void importHexAndVectorFiles(string rootName)
        {
            // Load vector file.
            FileInfo vecFile = new FileInfo(rootName + ".vector");
            if (vecFile.Exists)
            {
                TextReader vecRead = vecFile.OpenText();
                string vecStr = vecRead.ReadLine();
                debugVector = (uint)Int32.Parse(vecStr, System.Globalization.NumberStyles.AllowHexSpecifier);
                vecRead.Close();
            }
            // Clear device buffers.
            Pk2.ResetBuffers();

            // Load hex file
            switch (ImportExportHex.ImportHexFile(rootName))
            {
                case Constants.FileRead.success:
                    hexFilename = rootName;
                    break;

                default:
                    hexFilename = "";
                    Pk2.ResetBuffers();
                    break;
            }
        }

        // Update the firmware
        private void downloadNewFirmware(BackgroundWorker bw)
        {
            if (!Pk2BootLoader.ReadHexAndDownload(bw, hexFilename))
            {
                bw.ReportProgress(0, "Firmware download failed.");
                return;
            }

            // Verify new firmware
            if (!Pk2BootLoader.ReadHexAndVerify(bw, hexFilename))
            {
                bw.ReportProgress(0, "Firmware verification failed.");
                return;
            }

            // Write key 0x5555 at last memory location to tell bootloader FW loaded.
            if (!Pk2.BL_WriteFWLoadedKey())
            {
                bw.ReportProgress(0, "Firmware update failed.");
                return;
            }

            // Reset PICkit 2
            bw.ReportProgress(0, "Waiting for PICkit 2 to reset...");

            Pk2.BL_Reset();

            for (int i = 0; i < 50; i++)
            {
                bw.ReportProgress(2 * i, "Waiting for PICkit 2 to reset...");
                Thread.Sleep(100);
            }

            // Detect everything
            debugFW = false;
            detectEverything(bw, false);
        }

        // Detect a PICkit 2
        private void detectPICkit2(BackgroundWorker bw)
        {
            KONST.PICkit2USB detectResult;

            detectResult = Pk2.DetectPICkit2Device();

            pickitDetected = false;

            // PICkit 2 detected
            if (!debugFW && detectResult == KONST.PICkit2USB.found)
            {
                pickitDetected = true;
                needFirmwareUpdate = false;
                return;
            }

            // Something's wrong
            else
            {
                if (debugFW || detectResult == KONST.PICkit2USB.firmwareInvalid || detectResult == KONST.PICkit2USB.bootloader)
                {
                    needFirmwareUpdate = true;
                    hexFilename = KONST.FWFileName;
                    return;
                }
            }
        }

        // Detect a PIC18F part
        private void detectPIC()
        {
            picDetected = false;
            if (Pk2.DetectDevice(KONST.SEARCH_ALL_FAMILIES, true, false))
                picDetected = true;
            Pk2.SetVDDVoltage(5.0F);
        }

        // Snip off beginning of file path
        private string shortenHex(string fullPath)
        {
            if (fullPath.Length > 22)
            {
                return (fullPath.Substring(0, 3) + "..." + fullPath.Substring(fullPath.Length - 16));
            }
            return fullPath;
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            /*
            if (state != QP2State.ready)
                return;

            state = QP2State.busy;
            updateGUI();

            mBW.RunWorkerAsync(QP2Task.detect);
             * */
        }

        // Write a hex file
        private void deviceWrite(BackgroundWorker bw)
        {
            Pk2.VddOn();

            // Get OSCCAL if need be
            if (Pk2.DevFile.PartsList[Pk2.ActivePart].OSSCALSave)
            {
                Pk2.ReadOSSCAL();
                // put OSCCAL into part memory so it doesn't have to be written seperately.
                Pk2.DeviceBuffers.ProgramMemory[Pk2.DeviceBuffers.ProgramMemory.Length - 1] =
                    Pk2.DeviceBuffers.OSCCAL;
            }
            uint oscCal = Pk2.DeviceBuffers.OSCCAL;

            // Get BandGap if need be
            if (Pk2.DevFile.PartsList[Pk2.ActivePart].BandGapMask > 0)
            {
                Pk2.ReadBandGap();
            }
            uint bandGap = Pk2.DeviceBuffers.BandGap;

            // Erase Device First
            bool reWriteEE = false;

            bw.ReportProgress(0, "Erasing device...");

            Pk2.RunScript(KONST.PROG_ENTRY, 1);
            Pk2.RunScript(KONST.ERASE_CHIP, 1);
            Pk2.RunScript(KONST.PROG_EXIT, 1);

            if (Pk2.DevFile.PartsList[Pk2.ActivePart].WriteCfgOnErase)
            {
                Pk2.WriteConfigOutsideProgMem(false, false);
            }

            // Write the device
            if (hexFilename.Length == 0)
            {
                blankCheck(bw);
                return;
            }

            bw.ReportProgress(0, "Writing device...");
            Pk2.ResetBuffers();
            if (ImportExportHex.ImportHexFile(hexFilename) != Constants.FileRead.success)
            {
                bw.ReportProgress(0, "Couldn't read hex file.");
                return;
            }

            bool configInProgramSpace = false;

            // compute configration information.
            int configLocation = (int)Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigAddr /
                Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemHexBytes;
            int configWords = Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigWords;
            int endOfBuffer = Pk2.DeviceBuffers.ProgramMemory.Length;
            uint[] configBackups = new uint[configWords];  // use because DevicBuffers are masked & won't verify later
            if (configLocation < Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem)
            {// if config in program memory, set them to clear.
                configInProgramSpace = true;
                for (int cfg = configWords; cfg > 0; cfg--)
                {
                    configBackups[cfg - 1] = Pk2.DeviceBuffers.ProgramMemory[endOfBuffer - cfg];
                    Pk2.DeviceBuffers.ProgramMemory[endOfBuffer - cfg] =
                                Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue;
                }
            }
            endOfBuffer--;

            // Write Program Memory
            Pk2.RunScript(KONST.PROG_ENTRY, 1);

            if (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemWrPrepScript != 0)
            { // if prog mem address set script exists for this part
                Pk2.DownloadAddress3(0);
                Pk2.RunScript(KONST.PROGMEM_WR_PREP, 1);
            }
            int wordsPerWrite = Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemWrWords;
            int bytesPerWord = Pk2.DevFile.Families[Pk2.GetActiveFamily()].BytesPerLocation;
            int scriptRunsToUseDownload = KONST.DownLoadBufferSize /
                (wordsPerWrite * bytesPerWord);
            int wordsPerLoop = scriptRunsToUseDownload * wordsPerWrite;
            int wordsWritten = 0;

            // Find end of used memory
            endOfBuffer = Pk2.FindLastUsedInBuffer(Pk2.DeviceBuffers.ProgramMemory,
                                        Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue, endOfBuffer);
            // align end on next loop boundary                 
            int writes = (endOfBuffer + 1) / wordsPerLoop;
            if (((endOfBuffer + 1) % wordsPerLoop) > 0)
            {
                writes++;
            }
            endOfBuffer = writes * wordsPerLoop;

            int max = (int)endOfBuffer / wordsPerLoop;

            byte[] downloadBuffer = new byte[KONST.DownLoadBufferSize];

            do
            {
                int downloadIndex = 0;
                for (int word = 0; word < wordsPerLoop; word++)
                {
                    if (wordsWritten == endOfBuffer)
                    {
                        break; // for cases where ProgramMemSize%WordsPerLoop != 0
                    }
                    uint memWord = Pk2.DeviceBuffers.ProgramMemory[wordsWritten++];
                    if (Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemShift > 0)
                    {
                        memWord = memWord << 1;
                    }

                    downloadBuffer[downloadIndex++] = (byte)(memWord & 0xFF);

                    for (int bite = 1; bite < bytesPerWord; bite++)
                    {
                        memWord >>= 8;
                        downloadBuffer[downloadIndex++] = (byte)(memWord & 0xFF);
                    }

                }
                // download data
                int dataIndex = Pk2.DataClrAndDownload(downloadBuffer, 0);
                while (dataIndex < downloadIndex)
                {
                    dataIndex = Pk2.DataDownload(downloadBuffer, dataIndex);
                }

                Pk2.RunScript(KONST.PROGMEM_WR, scriptRunsToUseDownload);

                if ((wordsWritten == 0x8000) && (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemWrPrepScript != 0))
                { //PIC24 must update TBLPAG
                    Pk2.DownloadAddress3(0x10000);
                    Pk2.RunScript(KONST.PROGMEM_WR_PREP, 1);
                }

                bw.ReportProgress(100 * wordsWritten / endOfBuffer, "Writing device...");
            } while (wordsWritten < endOfBuffer);

            Pk2.RunScript(KONST.PROG_EXIT, 1);

            int verifyStop = endOfBuffer;

            if (configInProgramSpace)
            {// if config in program memory, restore prog memory to proper values.
                for (int cfg = configWords; cfg > 0; cfg--)
                {
                    Pk2.DeviceBuffers.ProgramMemory[Pk2.DeviceBuffers.ProgramMemory.Length - cfg]
                            = configBackups[cfg - 1];
                }
            }

            // Write debug vector
            byte[] data = {
                0x55, 0xAA, 0x55, 0xAA,
            };
            data[0] = (byte)((debugVector >> 1) & 0xFF);
            data[1] = 0xEF;
            data[2] = (byte)((debugVector >> 9) & 0xFF);
            data[3] = (byte)(0xF0 | ((debugVector >> 17) & 0x0F));

            // The PICkit 2 runs a sort of bytecode interpreter.
            byte[] script = {
                0xda, 0xa6, 0x8e,   // Set up EECON1
                0xda, 0xa6, 0x9c,

                0xda, 0x20, 0x0e,   // Set up TBLPTR
                0xda, 0xf8, 0x6e,
                0xda, 0x00, 0x0e,
                0xda, 0xf7, 0x6e,
                0xda, 0x28, 0x0e,
                0xda, 0xf6, 0x6e,

                0xee, 0x04, 0x0d,   // Send a byte
                0xf1,
                0xf1,
                
                0xee, 0x04, 0x0f,   // Send the last byte and start write
                0xf1,
                0xf1,

                0xee, 0x03, 0x00,   // Some other stuff which confuses me, frankly
                0xf3, 0x04,
                0xe7, 0x2f,
                0xf3, 0x00,
                0xf2, 0x00,
                0xf2, 0x00
            };

            Pk2.RunScript(KONST.PROG_ENTRY, 1);
            Pk2.DataClrAndDownload(data, 0);
            //Pk2.RunScript(KONST.USERID_WR, 1);
            Pk2.ExecuteScript(script);
            Pk2.RunScript(KONST.PROG_EXIT, 1);

            // Verify all but config (since hasn't been written as may contain code protection settings.
            if (configInProgramSpace)
            {// if config in program memory, don't verify configs.
                verifyStop = verifyStop - configWords;
            }
            bool verifySuccess = true;
            if (true)
            {
                verifySuccess = deviceVerify(bw, true, (verifyStop - 1), reWriteEE);
            }

            // WRITE CONFIGURATION
            if (verifySuccess)
            { // if we've failed verification, don't try to finish write
                // Write Configuration
                if ((configWords > 0) && (!configInProgramSpace))
                { // Write config words differently for any part where they are stored in program memory.

                    Pk2.WriteConfigOutsideProgMem(false, false);

                    // Verify Configuration
                    if (true)
                    {
                        if (verifyConfig(bw, configWords, configLocation))
                        {
                            bw.ReportProgress(100, "Programming successful.");
                        }
                    }
                }
                else if ((configWords > 0) && true)
                {  // for parts where config resides in program memory.
                    // program last memory block.

                    Pk2.RunScript(KONST.PROG_ENTRY, 1);
                    int lastBlock = Pk2.DeviceBuffers.ProgramMemory.Length -
                                    Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemWrWords;
                    if (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemWrPrepScript != 0)
                    { // if prog mem address set script exists for this part
                        Pk2.DownloadAddress3(lastBlock * Pk2.DevFile.Families[Pk2.GetActiveFamily()].AddressIncrement);
                        Pk2.RunScript(KONST.PROGMEM_WR_PREP, 1);
                    }
                    //byte[] downloadBuffer = new byte[KONST.DownLoadBufferSize];
                    int downloadIndex = 0;
                    for (int word = 0; word < Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemWrWords; word++)
                    {
                        uint memWord = Pk2.DeviceBuffers.ProgramMemory[lastBlock++];
                        if (Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemShift > 0)
                        {
                            memWord = memWord << 1;
                        }
                        downloadBuffer[downloadIndex++] = (byte)(memWord & 0xFF);
                        for (int bite = 1; bite < Pk2.DevFile.Families[Pk2.GetActiveFamily()].BytesPerLocation; bite++)
                        {
                            memWord >>= 8;
                            downloadBuffer[downloadIndex++] = (byte)(memWord & 0xFF);
                        }

                    }
                    // download data
                    int dataIndex = Pk2.DataClrAndDownload(downloadBuffer, 0);
                    while (dataIndex < downloadIndex)
                    {
                        dataIndex = Pk2.DataDownload(downloadBuffer, dataIndex);
                    }

                    Pk2.RunScript(KONST.PROGMEM_WR, 1);
                    Pk2.RunScript(KONST.PROG_EXIT, 1);
                    if (true)
                    {
                        bw.ReportProgress(100, "Programming successful.");
                    }

                }

                conditionalVDDOff();

            }
        }

        private bool deviceVerify(BackgroundWorker bw, bool writeVerify, int lastLocation, bool forceEEVerify)
        {
            bw.ReportProgress(0, "Verifying device...");

            int bytesPerWord = Pk2.DevFile.Families[Pk2.GetActiveFamily()].BytesPerLocation;

            Pk2.VddOn();

            byte[] upload_buffer = new byte[KONST.UploadBufferSize];

            // compute configration information.
            int configLocation = (int)Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigAddr /
                Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemHexBytes;
            int configWords = Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigWords;
            int endOfBuffer = Pk2.DeviceBuffers.ProgramMemory.Length - 1;
            if (writeVerify)
            { // unless it's a write-verify
                endOfBuffer = lastLocation;
            }

            //Verify Program Memory ----------------------------------------------------------------------------
            if (hexFilename.Length > 0)
            {

                Pk2.RunScript(KONST.PROG_ENTRY, 1);

                if ((Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrSetScript != 0)
                        && (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrBytes != 0))
                { // if prog mem address set script exists for this part
                    Pk2.DownloadAddress3(0);
                    Pk2.RunScript(KONST.PROGMEM_ADDRSET, 1);
                }


                int scriptRunsToFillUpload = KONST.UploadBufferSize /
                    (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemRdWords * bytesPerWord);
                int wordsPerLoop = scriptRunsToFillUpload * Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemRdWords;
                int wordsRead = 0;

                do
                {
                    Pk2.RunScriptUploadNoLen2(KONST.PROGMEM_RD, scriptRunsToFillUpload);
                    Array.Copy(Pk2.Usb_read_array, 1, upload_buffer, 0, KONST.USB_REPORTLENGTH);
                    Pk2.GetUpload();
                    Array.Copy(Pk2.Usb_read_array, 1, upload_buffer, KONST.USB_REPORTLENGTH, KONST.USB_REPORTLENGTH);
                    int uploadIndex = 0;
                    for (int word = 0; word < wordsPerLoop; word++)
                    {
                        int bite = 0;
                        uint memWord = (uint)upload_buffer[uploadIndex + bite++];
                        if (bite < bytesPerWord)
                        {
                            memWord |= (uint)upload_buffer[uploadIndex + bite++] << 8;
                        }
                        if (bite < bytesPerWord)
                        {
                            memWord |= (uint)upload_buffer[uploadIndex + bite++] << 16;
                        }
                        if (bite < bytesPerWord)
                        {
                            memWord |= (uint)upload_buffer[uploadIndex + bite++] << 24;
                        }
                        uploadIndex += bite;
                        // shift if necessary
                        if (Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemShift > 0)
                        {
                            memWord = (memWord >> 1) & Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue;
                        }

                        if (memWord != Pk2.DeviceBuffers.ProgramMemory[wordsRead++])
                        {
                            Pk2.RunScript(KONST.PROG_EXIT, 1);
                            conditionalVDDOff();
                            if (!writeVerify)
                            {
                                bw.ReportProgress(0, "Program memory verification failed.");
                            }
                            else
                            {
                                bw.ReportProgress(0, "Program memory programming failed.");
                            }

                            return false;
                        }
                        if ((wordsRead == 0x8000) && (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrSetScript != 0)
                                && (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrBytes != 0))
                        { //PIC24 must update TBLPAG
                            Pk2.DownloadAddress3(0x10000);
                            Pk2.RunScript(KONST.PROGMEM_ADDRSET, 1);
                            break;
                        }

                        if (wordsRead > endOfBuffer)
                        {
                            break; // for cases where ProgramMemSize%WordsPerLoop != 0
                        }
                    }
                    bw.ReportProgress(100 * wordsRead / endOfBuffer, "Verifying device...");
                } while (wordsRead < endOfBuffer);
                Pk2.RunScript(KONST.PROG_EXIT, 1);
            }

            //Verify debug vector ----------------------------------------------------------------------------------
            byte[] script = {
                0xda, 0x20, 0x0e,   // Set up TBLPTR
                0xda, 0xf8, 0x6e,
                0xda, 0x00, 0x0e,
                0xda, 0xf7, 0x6e,
                0xda, 0x28, 0x0e,
                0xda, 0xf6, 0x6e,

                0xee, 0x04, 0x09,   // Table read
                0xf2, 0x00,

                0xf0,               // Read into buffer

                0xe9, 0x06, 0x03    // Loop three more times
            };
            byte[] buf = new byte[4];
            for (int i = 0; i < 4; i++)
                buf[i] = 0x55;

            Pk2.RunScript(KONST.PROG_ENTRY, 1);
            //Pk2.RunScriptUploadNoLen(KONST.USERID_RD, 1);
            Pk2.ExecuteScriptUploadNoLen(script);

            Array.Copy(Pk2.Usb_read_array, 1, buf, 0, buf.Length);

            bool debugVectorCorrect = true;
            if (buf[0] != (byte)((debugVector >> 1) & 0xFF) ||
                buf[1] != 0xEF ||
                buf[2] != (byte)((debugVector >> 9) & 0xFF) ||
                buf[3] != (byte)(0xF0 | ((debugVector >> 17) & 0x0F)))
                debugVectorCorrect = false;

            if (debugVectorCorrect == false)
            {
                bw.ReportProgress(0, "Debug vector verification failed.");
                return false;
            }

            Pk2.RunScript(KONST.PROG_EXIT, 1);


            if (!writeVerify)
            {
                bw.ReportProgress(100, "Verification successful.");
                conditionalVDDOff();
            }
            return true;
        }

        private bool verifyConfig(BackgroundWorker bw, int configWords, int configLocation)
        {
            // Verify Configuration --------------------------------------------------------------------
            if ((configWords > 0) && (configLocation > Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem)
                    && true)
            { // Don't read config words for any part where they are stored in program memory.

                Pk2.RunScript(KONST.PROG_ENTRY, 1);
                Pk2.RunScript(KONST.CONFIG_RD, 1);
                Pk2.UploadData();
                Pk2.RunScript(KONST.PROG_EXIT, 1);
                int bufferIndex = 2;                    // report starts on index 1, which is #bytes uploaded.
                for (int word = 0; word < configWords; word++)
                {
                    uint config = (uint)Pk2.Usb_read_array[bufferIndex++];
                    config |= (uint)Pk2.Usb_read_array[bufferIndex++] << 8;
                    if (Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemShift > 0)
                    {
                        config = (config >> 1) & Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue;
                    }
                    config &= Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigMasks[word];
                    uint configExpected =
                        Pk2.DeviceBuffers.ConfigWords[word] & Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigMasks[word];

                    if (configExpected != config)
                    {
                        conditionalVDDOff();
                        bw.ReportProgress(0, "Configuration verification failed.");
                        return false;
                    }
                }
            }
            return true;
        }

        private void conditionalVDDOff()
        {
            if (!false)
            { // don't shut off if Vdd is "ON"
                Pk2.VddOff();
            }
        }

        private void eraseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (state != QP2State.ready)
                return;

            state = QP2State.busy;
            hexFilename = "";
            updateGUI();

            mBW.RunWorkerAsync(QP2Task.downloadHex);
        }

        private void detectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (state != QP2State.ready)
                return;

            state = QP2State.busy;
            updateGUI();

            mBW.RunWorkerAsync(QP2Task.detectAutoLoad);
        }

        private bool blankCheck(BackgroundWorker bw)
        {
            DeviceData blankDevice = new DeviceData(Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem,
                Pk2.DevFile.PartsList[Pk2.ActivePart].EEMem,
                Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigWords,
                Pk2.DevFile.PartsList[Pk2.ActivePart].UserIDWords,
                Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue,
                Pk2.DevFile.Families[Pk2.GetActiveFamily()].EEMemAddressIncrement,
                Pk2.DevFile.Families[Pk2.GetActiveFamily()].UserIDBytes,
                Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigBlank);

            // handle situation where configs are in program memory.
            int configLocation = (int)Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigAddr /
                Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemHexBytes;
            int configWords = Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigWords;
            if (configLocation < Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem)
                for (int i = 0; i < configWords; i++)
                {
                    uint template = blankDevice.ProgramMemory[configLocation + i] & 0xFFFF0000;
                    blankDevice.ProgramMemory[configLocation + i] =
                            (template | Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigBlank[i]);
                }

            bw.ReportProgress(0, "Verifying erasure...");

            Pk2.VddOn();

            byte[] upload_buffer = new byte[KONST.UploadBufferSize];

            //Check Program Memory ----------------------------------------------------------------------------
            {

                Pk2.RunScript(KONST.PROG_ENTRY, 1);

                if ((Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrSetScript != 0)
                        && (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrBytes != 0))
                { // if prog mem address set script exists for this part
                    Pk2.DownloadAddress3(0);
                    Pk2.RunScript(KONST.PROGMEM_ADDRSET, 1);
                }

                int bytesPerWord = Pk2.DevFile.Families[Pk2.GetActiveFamily()].BytesPerLocation;
                int scriptRunsToFillUpload = KONST.UploadBufferSize /
                    (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemRdWords * bytesPerWord);
                int wordsPerLoop = scriptRunsToFillUpload * Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemRdWords;
                int wordsRead = 0;

                do
                {
                    Pk2.RunScriptUploadNoLen2(KONST.PROGMEM_RD, scriptRunsToFillUpload);
                    Array.Copy(Pk2.Usb_read_array, 1, upload_buffer, 0, KONST.USB_REPORTLENGTH);
                    Pk2.GetUpload();
                    Array.Copy(Pk2.Usb_read_array, 1, upload_buffer, KONST.USB_REPORTLENGTH, KONST.USB_REPORTLENGTH);
                    int uploadIndex = 0;
                    for (int word = 0; word < wordsPerLoop; word++)
                    {
                        int bite = 0;
                        uint memWord = (uint)upload_buffer[uploadIndex + bite++];
                        if (bite < bytesPerWord)
                        {
                            memWord |= (uint)upload_buffer[uploadIndex + bite++] << 8;
                        }
                        if (bite < bytesPerWord)
                        {
                            memWord |= (uint)upload_buffer[uploadIndex + bite++] << 16;
                        }
                        if (bite < bytesPerWord)
                        {
                            memWord |= (uint)upload_buffer[uploadIndex + bite++] << 24;
                        }
                        uploadIndex += bite;
                        // shift if necessary
                        if (Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemShift > 0)
                        {
                            memWord = (memWord >> 1) & Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue;
                        }

                        // if OSCCAL save, force last word to be blank
                        if ((Pk2.DevFile.PartsList[Pk2.ActivePart].OSSCALSave)
                                && wordsRead == (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem - 1))
                        {
                            memWord = Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue;
                        }
                        if (memWord != blankDevice.ProgramMemory[wordsRead++])
                        {
                            Pk2.RunScript(KONST.PROG_EXIT, 1);
                            conditionalVDDOff();
                            bw.ReportProgress(0, "Program memory erasure failed.");
                            return false;
                        }

                        if (wordsRead == Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem)
                        {
                            break; // for cases where ProgramMemSize%WordsPerLoop != 0
                        }
                        if ((wordsRead == 0x8000) && (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrSetScript != 0)
                                    && (Pk2.DevFile.PartsList[Pk2.ActivePart].ProgMemAddrBytes != 0))
                        { //PIC24 must update TBLPAG
                            Pk2.DownloadAddress3(0x10000);
                            Pk2.RunScript(KONST.PROGMEM_ADDRSET, 1);
                            break;
                        }
                    }
                    bw.ReportProgress(100*wordsRead / (int)Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem,
                        "Verifying erasure...");
                } while (wordsRead < Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem);
                Pk2.RunScript(KONST.PROG_EXIT, 1);
            }

            //Check debug vector ----------------------------------------------------------------------------------
            byte[] script = {
                0xda, 0x20, 0x0e,   // Set up TBLPTR
                0xda, 0xf8, 0x6e,
                0xda, 0x00, 0x0e,
                0xda, 0xf7, 0x6e,
                0xda, 0x28, 0x0e,
                0xda, 0xf6, 0x6e,

                0xee, 0x04, 0x09,   // Table read
                0xf2, 0x00,

                0xf0,               // Read into buffer

                0xe9, 0x06, 0x03    // Loop three more times
            };
            byte[] buf = new byte[4];
            for (int i = 0; i < 4; i++)
                buf[i] = 0x55;

            Pk2.RunScript(KONST.PROG_ENTRY, 1);
            Pk2.ExecuteScriptUploadNoLen(script);

            Array.Copy(Pk2.Usb_read_array, 1, buf, 0, buf.Length);

            if (buf[0] != 0xff || buf[1] != 0xff || buf[2] != 0xff || buf[3] != 0xff)
            {
                bw.ReportProgress(0, "Debug vector erasure failed.");
                return false;
            }

            // Blank Check Configuration --------------------------------------------------------------------
            if ((configWords > 0) && (configLocation > Pk2.DevFile.PartsList[Pk2.ActivePart].ProgramMem))
            { // Don't read config words for any part where they are stored in program memory.
                Pk2.RunScript(KONST.PROG_ENTRY, 1);
                Pk2.RunScript(KONST.CONFIG_RD, 1);
                Pk2.UploadData();
                Pk2.RunScript(KONST.PROG_EXIT, 1);
                int bufferIndex = 2;                    // report starts on index 1, which is #bytes uploaded.
                for (int word = 0; word < configWords; word++)
                {
                    uint config = (uint)Pk2.Usb_read_array[bufferIndex++];
                    config |= (uint)Pk2.Usb_read_array[bufferIndex++] << 8;
                    if (Pk2.DevFile.Families[Pk2.GetActiveFamily()].ProgMemShift > 0)
                    {
                        config = (config >> 1) & Pk2.DevFile.Families[Pk2.GetActiveFamily()].BlankValue;
                    }
                    config &= Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigMasks[word];
                    int configBlank = Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigMasks[word]
                                       & Pk2.DevFile.PartsList[Pk2.ActivePart].ConfigBlank[word];
                    if (configBlank != config)
                    {
                        bw.ReportProgress(0, "Configuration erasure failed.");
                        return false;
                    }
                }
            }


            Pk2.RunScript(KONST.PROG_EXIT, 1);
            conditionalVDDOff();

            bw.ReportProgress(100, "Erasure successful.");
            return true;
        }
    }
}